/*
 * Copyright (c) 2005-2007 Extreme! Lab, Indiana University. All rights reserved.
 *
 * This software is open source. See the bottom of this file for the license.
 *
 * $Id: GenericInvoker.java,v 1.18 2007/04/24 01:14:04 sshirasu Exp $
 */

package edu.indiana.extreme.xbaya.jython.lib;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import javax.xml.namespace.QName;

import org.xmlpull.v1.builder.XmlElement;

import xsul.lead.LeadContextHeader;
import xsul.wsdl.WsdlDefinitions;
import xsul.wsdl.WsdlException;
import xsul.wsdl.WsdlResolver;
import xsul.wsif.WSIFMessage;
import xsul.xhandler_soap_sticky_header.StickySoapHeaderHandler;
import xsul.xwsif_runtime.WSIFClient;
import xsul5.MLogger;
import edu.indiana.extreme.xbaya.XBayaException;
import edu.indiana.extreme.xbaya.XBayaRuntimeException;
import edu.indiana.extreme.xbaya.jython.lib.invoker.Invoker;
import edu.indiana.extreme.xbaya.jython.lib.invoker.InvokerFactory;
import edu.indiana.extreme.xbaya.lead.LeadContextHeaderHelper;
import edu.indiana.extreme.xbaya.util.XMLUtil;

/**
 * The InvokerWithNotification class invokes a Web service.
 * 
 * @author Satoshi Shirasuna
 */
public class GenericInvoker {

    private static final MLogger logger = MLogger.getLogger();

    private String nodeID;

    private QName portTypeQName;

    private String wsdlLocation;

    private String messageBoxURL;

    private String resourceCatalogURL;

    private String gfacURL;

    private Invoker invoker;

    private Future<Boolean> result;

    private ServiceNotificationSender notifier;

    /**
     * used for notification
     */
    private List<Object> inputValues = new ArrayList<Object>();

    /**
     * used for notification
     */
    private List<String> inputNames = new ArrayList<String>();

    boolean failerSent;

    private String gfacRegistryURL;

    /**
     * Creates an InvokerWithNotification.
     * 
     * @param portTypeQName
     * 
     * @param wsdlLocation
     *            The URL of WSDL of the service to invoke
     * @param nodeID
     *            The ID of the service
     * @param resourceCatalogURL
     * @param gfacURL
     *            The URL of GFac service.
     * @param notifier
     *            The notification sender
     */
    public GenericInvoker(QName portTypeQName, String wsdlLocation,
            String nodeID, String resourceCatalogURL, String gfacURL,
            NotificationSender notifier) {
        this(portTypeQName, wsdlLocation, nodeID, resourceCatalogURL, gfacURL,
                null, notifier);
    }

    /**
     * Creates an InvokerWithNotification.
     * 
     * @param portTypeQName
     * 
     * @param wsdlLocation
     *            The URL of WSDL of the service to invoke
     * @param nodeID
     *            The ID of the service
     * @param resourceCatalogURL
     * @param gfacRegistryURL
     * @param gfacURL
     *            The URL of GFac service.
     * @param notifier
     *            The notification sender
     */
    public GenericInvoker(QName portTypeQName, String wsdlLocation,
            String nodeID, String resourceCatalogURL, String gfacRegistryURL,
            String gfacURL, NotificationSender notifier) {
        this(portTypeQName, wsdlLocation, nodeID, null, resourceCatalogURL,
                gfacRegistryURL, gfacURL, notifier);
    }

    /**
     * Creates an InvokerWithNotification.
     * 
     * @param portTypeQName
     * 
     * @param wsdlLocation
     *            The URL of WSDL of the service to invoke
     * @param nodeID
     *            The ID of the service
     * @param messageBoxURL
     * @param resourceCatalogURL
     * @param gfacRegistryURL
     * @param gfacURL
     *            The URL of GFac service.
     * @param notifier
     *            The notification sender
     */
    public GenericInvoker(QName portTypeQName, String wsdlLocation,
            String nodeID, String messageBoxURL, String resourceCatalogURL,
            String gfacRegistryURL, String gfacURL, NotificationSender notifier) {
        logger.entering(new Object[] { portTypeQName, wsdlLocation, nodeID,
                notifier });
        this.nodeID = nodeID;
        this.portTypeQName = portTypeQName;
        this.wsdlLocation = wsdlLocation;
        this.messageBoxURL = messageBoxURL;
        this.resourceCatalogURL = resourceCatalogURL;
        this.gfacURL = gfacURL;
        this.gfacRegistryURL = gfacRegistryURL;
        this.notifier = notifier.createServiceNotificationSender(nodeID);

        this.failerSent = false;
    }

    /**
     * Sets up the invoker.
     * 
     * @throws XBayaException
     */
    public void setup() throws XBayaException {
        try {
            WsdlDefinitions definitions = null;
            if (this.wsdlLocation != null && !this.wsdlLocation.equals("")) {
                definitions = WsdlResolver.getInstance().loadWsdl(
                        new File(".").toURI(), new URI(this.wsdlLocation));
            }

            // Create Invoker
            this.invoker = InvokerFactory.createInvoker(this.portTypeQName,
                    definitions, this.resourceCatalogURL, this.gfacRegistryURL,
                    this.gfacURL, this.messageBoxURL);

            this.invoker.setup();

            // Set LEAD context header.
            WSIFClient client = this.invoker.getClient();

            LeadContextHeaderHelper leadContextHelper = new LeadContextHeaderHelper();
            LeadContextHeader leadContext = leadContextHelper
                    .getLeadContextHeader();
            leadContext.setEventSink(this.notifier.getEventSink());
            StickySoapHeaderHandler handler = new StickySoapHeaderHandler(
                    "use-lead-header", leadContext);
            client.addHandler(handler);
            leadContext.setServiceId(this.nodeID);
            leadContext.setNodeId(this.nodeID);
            leadContext.setWorkflowId(this.notifier.getWorkflowID());
            leadContext.setTimeStep("1");

            // Get the concrete WSDL from invoker.setup() and set it to the
            // notifier.
            String name = this.portTypeQName.getLocalPart();
            this.notifier.setServiceID(name);

            logger.exiting();
        } catch (XBayaException e) {
            logger.caught(e);
            // An appropriate message has been set in the exception.
            this.notifier.invocationFailed(e.getMessage(), e);
            throw e;
        } catch (URISyntaxException e) {
            logger.caught(e);
            String message = "The location of the WSDL has to be a valid URL or file path: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (WsdlException e) {
            logger.caught(e);
            String message = "Error in processing the WSDL: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "Error in processing the WSDL: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }

    /**
     * Sets the operation name to invoke.
     * 
     * @param operationName
     *            The name of the operation
     * @throws XBayaException
     */
    public void setOperation(String operationName) throws XBayaException {

        logger.entering(new Object[] { operationName });

        try {
            this.invoker.setOperation(operationName);
            logger.exiting();
        } catch (XBayaException e) {
            logger.caught(e);
            // An appropriate message has been set in the exception.
            this.notifier.invocationFailed(e.getMessage(), e);
            throw e;
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "The WSDL does not conform to the invoking service: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }

    /**
     * Sets an input parameter
     * 
     * @param name
     *            The name of the input parameter
     * @param value
     *            The value of the input parameter
     * @throws XBayaException
     */
    public void setInput(String name, Object value) throws XBayaException {
        logger.entering(new Object[] { name, value });
        try {
            if (value instanceof XmlElement) {
                logger.finest("value: "
                        + XMLUtil.xmlElementToString((XmlElement) value));
            }
            this.inputNames.add(name);
            this.inputValues.add(value);
            this.invoker.setInput(name, value);
            logger.exiting();
        } catch (XBayaException e) {
            logger.caught(e);
            // An appropriate message has been set in the exception.
            this.notifier.invocationFailed(e.getMessage(), e);
            throw e;
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "Error in setting an input. name: " + name
                    + " value: " + value;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }

    /**
     * Invokes the service. This is a non-blocking call.
     * 
     * @throws XBayaException
     */
    public synchronized void invoke() throws XBayaException {

        logger.entering();

        try {
            WSIFMessage inputMessage = this.invoker.getInputs();
            logger.finest("inputMessage: "
                    + XMLUtil.xmlElementToString((XmlElement) inputMessage));
            this.notifier.invokingService(inputMessage);

            ExecutorService executor = Executors.newSingleThreadExecutor();
            this.result = executor.submit(new Callable<Boolean>() {
                @SuppressWarnings("boxing")
                public Boolean call() {
                    try {
                        boolean success = GenericInvoker.this.invoker.invoke();
                        if (success) {
                            // Send notification
                            WSIFMessage outputMessage = GenericInvoker.this.invoker
                                    .getOutputs();
                            // An implementation of WSIFMessage,
                            // WSIFMessageElement, implements toString(), which
                            // serialize the message XML.
                            logger.finest("outputMessage: " + outputMessage);
                            GenericInvoker.this.notifier
                                    .serviceFinished(outputMessage);
                        } else {
                            WSIFMessage faultMessage = GenericInvoker.this.invoker
                                    .getFault();
                            // An implementation of WSIFMessage,
                            // WSIFMessageElement, implements toString(), which
                            // serialize the message XML.
                            logger.finest("received fault: " + faultMessage);
                            GenericInvoker.this.notifier
                                    .receivedFault(faultMessage);
                            GenericInvoker.this.failerSent = true;
                        }
                        return success;
                    } catch (XBayaException e) {
                        logger.caught(e);
                        // An appropriate message has been set in the exception.
                        GenericInvoker.this.notifier.invocationFailed(e
                                .getMessage(), e);
                        GenericInvoker.this.failerSent = true;
                        throw new XBayaRuntimeException(e);
                    } catch (RuntimeException e) {
                        logger.caught(e);
                        String message = "Error in invoking a service: "
                                + GenericInvoker.this.wsdlLocation;
                        GenericInvoker.this.notifier.invocationFailed(message,
                                e);
                        GenericInvoker.this.failerSent = true;
                        throw e;
                    }
                }
            });

            // Kill the thread inside of executor. This is necessary for Jython
            // script to finish.
            executor.shutdown();

            // Let other threads know that job has been submitted.
            notifyAll();

            // Check if the invocation itself fails. This happens immediately.
            try {
                this.result.get(100, TimeUnit.MILLISECONDS);
            } catch (InterruptedException e) {
                logger.caught(e);
            } catch (TimeoutException e) {
                // The job is probably running fine.
                // The normal case.
                return;
            } catch (ExecutionException e) {
                // The service-failed notification should have been sent
                // already.
                logger.caught(e);
                String message = "Error in invoking a service: "
                        + this.wsdlLocation;
                throw new XBayaException(message, e);
            }

            logger.exiting();
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "Error in invoking a service: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }

    /**
     * Waits until the execution finishes.
     * 
     * @throws XBayaException
     */
    @SuppressWarnings("boxing")
    public synchronized void waitToFinish() throws XBayaException {
        logger.entering();
        try {
            while (this.result == null) {
                // The job is not submitted yet.
                try {
                    wait();
                } catch (InterruptedException e) {
                    logger.caught(e);
                }
            }
            // Wait for the job to finish.
            Boolean success = this.result.get();
            if (success == false) {
                WSIFMessage faultMessage = this.invoker.getFault();
                String message = "Error in a service: ";
                // An implementation of WSIFMessage,
                // WSIFMessageElement, implements toString(), which
                // serialize the message XML.
                message += faultMessage.toString();
                throw new XBayaException(message);
            }
            logger.exiting();
        } catch (InterruptedException e) {
            logger.caught(e);
        } catch (ExecutionException e) {
            // The service-failed notification should have been sent already.
            logger.caught(e);
            String message = "Error in invoking a service: "
                    + this.wsdlLocation;
            throw new XBayaException(message, e);
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "Error while waiting for a service to finish: "
                    + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }

    /**
     * Returns the output of a specified name. This method blocks until the
     * execution finishes.
     * 
     * @param name
     *            The name of the output parameter
     * @return The value of the output
     * @throws XBayaException
     */
    public Object getOutput(String name) throws XBayaException {
        logger.entering(new Object[] { name });
        try {
            waitToFinish();
            Object output = this.invoker.getOutput(name);
            if (output instanceof XmlElement) {
                logger.finest("output: "
                        + XMLUtil.xmlElementToString((XmlElement) output));
            }
            logger.exiting(output);
            return output;
        } catch (XBayaException e) {
            logger.caught(e);
            // An appropriate message has been set in the exception.
            if (!this.failerSent) {
                this.notifier.invocationFailed(e.getMessage(), e);
            }
            throw e;
        } catch (RuntimeException e) {
            logger.caught(e);
            String message = "Error while waiting for a output: " + name;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        } catch (Error e) {
            logger.caught(e);
            String message = "Unexpected error: " + this.wsdlLocation;
            this.notifier.invocationFailed(message, e);
            throw new XBayaException(message, e);
        }
    }
}

/*
 * Indiana University Extreme! Lab Software License, Version 1.2
 * 
 * Copyright (c) 2005-2007 The Trustees of Indiana University. All rights
 * reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1) All redistributions of source code must retain the above copyright notice,
 * the list of authors in the original source code, this list of conditions and
 * the disclaimer listed in this license;
 * 
 * 2) All redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the disclaimer listed in this license in
 * the documentation and/or other materials provided with the distribution;
 * 
 * 3) Any documentation included with all redistributions must include the
 * following acknowledgement:
 * 
 * "This product includes software developed by the Indiana University Extreme!
 * Lab. For further information please visit http://www.extreme.indiana.edu/"
 * 
 * Alternatively, this acknowledgment may appear in the software itself, and
 * wherever such third-party acknowledgments normally appear.
 * 
 * 4) The name "Indiana University" or "Indiana University Extreme! Lab" shall
 * not be used to endorse or promote products derived from this software without
 * prior written permission from Indiana University. For written permission,
 * please contact http://www.extreme.indiana.edu/.
 * 
 * 5) Products derived from this software may not use "Indiana University" name
 * nor may "Indiana University" appear in their name, without prior written
 * permission of the Indiana University.
 * 
 * Indiana University provides no reassurances that the source code provided
 * does not infringe the patent or any other intellectual property rights of any
 * other entity. Indiana University disclaims any liability to any recipient for
 * claims brought by any other entity based on infringement of intellectual
 * property rights or otherwise.
 * 
 * LICENSEE UNDERSTANDS THAT SOFTWARE IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. INDIANA UNIVERSITY GIVES
 * NO WARRANTIES AND MAKES NO REPRESENTATION THAT SOFTWARE IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER PROPRIETARY RIGHTS.
 * INDIANA UNIVERSITY MAKES NO WARRANTIES THAT SOFTWARE IS FREE FROM "BUGS",
 * "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS", OR OTHER HARMFUL CODE.
 * LICENSEE ASSUMES THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF INFORMATION
 * GENERATED USING SOFTWARE.
 */
